﻿Option Strict On

Imports Nbis
Imports System.Text
Imports System.IO
Imports System.Text.RegularExpressions

<TestClass()>
Public Class NbisNetTests

    <TestMethod()>
    Public Sub MinutiaeCompare_cmp00001cmp00011_expected12()
        'Detect Minutiae
        Dim Probe As Minutia() = Nbis.DetectMinutiae.FromBitmap(Nbis.Wsq.FromWsqToBitmap(My.Resources.cmp00001), 500)
        Dim Gallery As Minutia() = Nbis.DetectMinutiae.FromBitmap(Nbis.Wsq.FromWsqToBitmap(My.Resources.cmp00011), 500)

        'Bozorth3 Minutiae Matcher
        Dim Result As Integer = Nbis.Matcher.Compare(Probe, Gallery)
        'System.Console.WriteLine("Result: {0}", Result)

        'Test compares an image to another image-resulting matching minutiae should be 12 (for cmp00001)
        Assert.AreEqual(12, Result, "Matcher.Compare not accurate")

    End Sub

    <TestMethod()>
    Public Sub NfiqFromFile_cmp00013_ExpectedThree()
        'Return NFIQ score-from filename input
        Dim NfiqResult As Integer = Nbis.Nfiq.FromWsq(My.Resources.cmp00013)

        Assert.AreEqual(3, NfiqResult, "NFIQ score not accurate")
    End Sub

    <TestMethod()>
    Public Sub NfiqFromBitmap_cmp00013_expectedThree()
        'Return NFIQ score-from bitmap file input
        Dim NfiqResult As Integer = Nbis.Nfiq.FromBitmap(Nbis.Wsq.FromWsqToBitmap(My.Resources.cmp00013), 500)
        Assert.AreEqual(3, NfiqResult, "NFIQ score not accurate")
    End Sub

    <TestMethod()>
    Public Sub NfiqFromWsq_cmp00013_expectedThree()
        'Return NFIQ score-from wsq file input
        Dim NfiqResult As Integer = Nbis.Nfiq.FromWsq(My.Resources.cmp00013)
        Assert.AreEqual(3, NfiqResult, "NFIQ score not accurate")
    End Sub

    <TestMethod()>
    Public Sub DetectMinutiaFromBitmap_compareToCorrectValues_expectedValuesWithinRange()

        'Detect Minutiae fromBitmap (verifies that minutiae results are the same for the same bitmap image
        Dim Probe As Minutia() = Nbis.DetectMinutiae.FromBitmap(Nbis.Wsq.FromWsqToBitmap(My.Resources.a011p_03), 500)
        System.Console.WriteLine("Checking probe values.. ")

        Dim o As Byte() = My.Resources.a011p_031  'reads the .xyt file in as an array of bytes. cannot convert an array of bytes directly to a streamreader object
        Dim ms As New MemoryStream(o)
        Dim tmpstream As New StreamReader(ms)
        Dim lines As String()
        Dim aline As String()

        lines = tmpstream.ReadToEnd.Split(CChar(vbLf)) 'read the xytq file

        If lines(lines.GetUpperBound(0)).Length = 0 Then 'if the last element in the lines array has no data delete it (redim the array)
            ReDim Preserve lines(lines.GetUpperBound(0) - 1)
        End If

        Dim minutiaFromFile(lines.GetUpperBound(0)) As Minutia
        Dim counter As Integer = 0

        System.Console.WriteLine("Against datafile values.. ")

        For x = 0 To lines.GetUpperBound(0)
            aline = lines(x).Split(CChar(" "))
            minutiaFromFile(x) = New Minutia()
            minutiaFromFile(x).X = CInt(aline(0))
            minutiaFromFile(x).Y = CInt(aline(1))
            minutiaFromFile(x).Direction = CInt(aline(2))
            minutiaFromFile(x).Quality = CInt(aline(3))

            counter += 1
        Next

        For i As Integer = 0 To minutiaFromFile.GetUpperBound(0)
            Assert.AreEqual(minutiaFromFile(i).X, Probe(i).X, "minutia x values not equal")
            Assert.AreEqual(minutiaFromFile(i).Y, Probe(i).Y, "minutia y values not equal")
            Assert.AreEqual(minutiaFromFile(i).Direction, Probe(i).Direction, "minutia direction values not equal")
            Assert.AreEqual(minutiaFromFile(i).Quality, Probe(i).Quality, "minutia quality values not equal")
        Next
    End Sub

    <TestMethod()>
    Public Sub Segmentation_CompareToCorrectSegmentedValues_expectedWithinRange()

        'test verifies values generated by Nbis.Segmentation algorithm (run on a wsq file) are the same (within legal limits) as 
        'those groundtruthed in a prerecorded text file for that image.

        Dim Parameters As New Nbis.SegmenterParameters 'parameters needed for segmentation algorithm
        Parameters.TargetFingers = Finger.RightSlap
        Parameters.PermitRotation = True
        Parameters.IsInked = True

        Dim Result As Nbis.Segment() = Nbis.Segmentation.FromBitmap(Nbis.Wsq.FromWsqToBitmap(My.Resources.a011_13), Parameters)
        System.Console.WriteLine("image file a011_13 output values: ")
        For i As Integer = 0 To Result.GetUpperBound(0) '3 needs to be determined by the file contents not hardcoded.
            System.Console.WriteLine(Result(i).Width & " " & Result(i).Height & " " & Result(i).OriginX & " " & Result(i).OriginY & " " & Result(i).Rotation)
        Next

        Dim line As String
        Dim aline() As String
        Dim o As Byte() = My.Resources.a011_131  'reads the .xyt file in as an array of bytes. cannot convert an array of bytes directly to a streamreader object
        Dim ms As New MemoryStream(o)
        Dim tmpstream As New StreamReader(ms)

        Dim x As Integer = 0
        Dim datafile(Result.GetUpperBound(0)) As Segment
        Dim degreeRotation As Single

        System.Console.WriteLine("groundtruth values: ")
        Do Until tmpstream.EndOfStream
            line = tmpstream.ReadLine()
            aline = line.Split(CChar(" "))
            datafile(x) = New Nbis.Segment()
            datafile(x).Width = CInt(aline(0))
            datafile(x).Height = CInt(aline(1))
            datafile(x).OriginX = CInt(aline(2))
            datafile(x).OriginY = CInt(aline(3))
            datafile(x).Rotation = Single.Parse(aline(4))
            System.Console.WriteLine(datafile(x).Width & " " & datafile(x).Height & " " & datafile(x).OriginX & " " & datafile(x).OriginY & " " & datafile(x).Rotation)
            x += 1
        Loop

        For j As Integer = 0 To datafile.GetUpperBound(0)
            Assert.AreEqual(Result(j).Width, datafile(j).Width, "Width values are not equal")
            Assert.AreEqual(Result(j).Height, datafile(j).Height, "Width values are not equal")

            Assert.AreEqual(Result(j).OriginX, datafile(j).OriginX, "originX value not equal")
            Assert.AreEqual(Result(j).OriginY, datafile(j).OriginY, "originY value not equal")
            degreeRotation = CSng(Math.Round(Result(j).Rotation * 180 / Math.PI, 1)) 'converts radians to degrees
            Assert.AreEqual(degreeRotation, datafile(j).Rotation, "Rotation value not equal")
        Next
    End Sub
End Class